#ifndef {{ NAME_SNAKE_UPPER }}_{{ NAME_SNAKE_UPPER }}_HPP
#define {{ NAME_SNAKE_UPPER }}_{{ NAME_SNAKE_UPPER }}_HPP

#include <{{ name }}/export.hpp>

namespace {{ name_snake_lower }} {

{{ NAME_SNAKE_UPPER }}_EXPORT void {{ name_snake_lower }}();

}

#endif
